#ifndef BOSON_AST_H
#define BOSON_AST_H

#include <stdbool.h>
#include <stddef.h>
#include <stdint.h>

#include "lexer.h"

enum ast_type {
	AST_NONE,
	AST_ASSIGNMENT,
	AST_CONDITION,
	AST_OR,
	AST_AND,
	AST_EQUALITY,
	AST_RELATION,
	AST_ADDITION,
	AST_MULTIPLICATION,
	AST_UNARY,
	AST_SUBSCRIPT,
	AST_FUNCTION,
	AST_METHOD,
	AST_IDENTIFIER,
	AST_STRING,
	AST_ARRAY,
	AST_BOOL,
	AST_NUMBER,
};

struct ast_expression;

struct ast_bool {
	bool value;
};

enum ast_number_base {
	NUMBER_BIN = 2,
	NUMBER_OCT = 8,
	NUMBER_DEC = 10,
	NUMBER_HEX = 16,
};

struct ast_number{
	int64_t value;
	enum ast_number_base base;
};

struct ast_identifier {
	char *data;
	size_t n;
};

struct ast_string {
	char *data;
	size_t n;
};

struct ast_identifier_list {
	struct ast_identifier **identifiers;
	size_t n;
};

struct ast_expression_list {
	struct ast_expression **values;
	size_t n;
};

struct ast_keyword_list {
	struct ast_identifier **keys;
	struct ast_expression **values;
	size_t n;
};

struct ast_arguments {
	struct ast_expression_list *position;
	struct ast_keyword_list *keyword;
};

struct ast_subscript {
	struct ast_expression *left;
	struct ast_expression *right;
};

struct ast_function {
	struct ast_identifier *left;
	struct ast_arguments *right;
};

struct ast_method {
	struct ast_expression *left;
	struct ast_function *right;
};

enum ast_unary_op {
	UNARY_NOT,
	UNARY_PLUS,
	UNARY_MINUS,
};

struct ast_unary {
	enum ast_unary_op op;
	struct ast_expression *right;
};

enum ast_multiplication_op {
	MULTIPLICATION_STAR = 1 << 0,
	MULTIPLICATION_SLASH = 1 << 1,
	MULTIPLICATION_MOD = 1 << 2,
};

struct ast_multiplication {
	struct ast_expression *left;
	enum ast_multiplication_op op;
	struct ast_expression *right;
};

enum ast_addition_op {
	ADDITION_PLUS = 1 << 0,
	ADDITION_MINUS = 1 << 1,
};

struct ast_addition {
	struct ast_expression *left;
	enum ast_addition_op op;
	struct ast_expression *right;
};

enum ast_relation_op {
	RELATION_GT = 1 << 0,
	RELATION_LT = 1 << 1,
	RELATION_GEQ = 1 << 2,
	RELATION_LEQ = 1 << 3,
	RELATION_IN = 1 << 4,
	RELATION_NIN = 1 << 5,
};

struct ast_relation {
	struct ast_expression *left;
	enum ast_relation_op op;
	struct ast_expression *right;
};

enum ast_equality_op {
	EQUALITY_EQ = 1 << 0,
	EQUALITY_NEQ = 1 << 1
};

struct ast_equality {
	struct ast_expression *equality;
	enum ast_equality_op op;
	struct ast_expression *relation;
};

struct ast_and {
	struct ast_expression *left;
	struct ast_expression *right;
};

struct ast_or {
	struct ast_expression *left;
	struct ast_expression *right;
};

struct ast_condition {
	struct ast_expression *cond;
	struct ast_expression *left;
	struct ast_expression *right;
};

enum ast_assignment_op {
	ASSIGNMENT_ASSIGN = 1 << 0,
	ASSIGNMENT_STAREQ = 1 << 1,
	ASSIGNMENT_SLASHEQ = 1 << 2,
	ASSIGNMENT_MODEQ = 1 << 3,
	ASSIGNMENT_PLUSEQ = 1 << 4,
	ASSIGNMENT_MINEQ = 1 << 5,
};

struct ast_assignment {
	struct ast_identifier *left;
	enum ast_assignment_op op;
	struct ast_expression *right;
};

struct ast_expression {
	enum ast_type type;
	union {
		struct ast_assignment *assignment;
		struct ast_condition *condition;
		struct ast_or *or;
		struct ast_and *and;
		struct ast_equality *equality;
		struct ast_relation *relation;
		struct ast_addition *addition;
		struct ast_multiplication *multiplication;
		struct ast_unary *unary;
		struct ast_subscript *subscript;
		struct ast_function *function;
		struct ast_method *method;
		struct ast_identifier *identifier;
		struct ast_string *string;
		struct ast_expression_list *array;
		struct ast_bool *boolean;
		struct ast_number *number;
	};
	struct location loc;
};

struct ast_selection {
	int dummy;
};

struct ast_iteration {
	int dummy;
};

enum ast_statement_type {
	STATEMENT_EXPRESSION = 1 << 0,
	STATEMENT_SELECTION = 1 << 1,
	STATEMENT_ITERATION = 1 << 2,
};

struct ast_statement {
	enum ast_statement_type type;
	union {
		struct ast_expression *expression;
		struct ast_selection *selection;
		struct ast_iteration *iteration;
	};
	struct ast_statement *next;
	struct ast_statement *prev;
};

const char *ast_expression_str(struct ast_expression *);

#endif // BOSON_AST_H
