#define _XOPEN_SOURCE 700

#include <ctype.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>

#include "lexer.h"
#include "log.h"
#include "util.h"

// Order must match lexer.h
static const char *tokens[] = {
	// Keywords
	[T_AND] = "and",
	[T_BREAK] = "break",
	[T_CONTINUE] = "continue",
	[T_ELIF] = "elif",
	[T_ELSE] = "else",
	[T_ENDFOREACH] = "endforeach",
	[T_ENDIF] = "endif",
	[T_FALSE] = "false",
	[T_FOREACH] = "foreach",
	[T_IF] = "if",
	[T_IN] = "in",
	[T_NOT] = "not",
	[T_OR] = "or",
	[T_TRUE] = "true",

	// Operators
	[T_ASSIGN] = "=",
	[T_COLON] = ":",
	[T_COMMA] = ",",
	[T_DOT] = ".",
	[T_EQ] = "==",
	[T_GEQ] = ">=",
	[T_GT] = ">",
	[T_LBRACK] = "[",
	[T_LCURL] = "{",
	[T_LEQ] = "<=",
	[T_LPAREN] = "(",
	[T_LT] = "<",
	[T_MINEQ] = "-=",
	[T_MINUS] = "-",
	[T_MODEQ] = "%=",
	[T_MODULO] = "%",
	[T_NEQ] = "!=",
	[T_PLUS] = "+",
	[T_PLUSEQ] = "+=",
	[T_QUESTION] = "?",
	[T_RBRACK] = "]",
	[T_RCURL] = "}",
	[T_RPAREN] = ")",
	[T_SLASH] = "/",
	[T_SLASHEQ] = "/=",
	[T_STAR] = "*",
	[T_STAREQ] = "*=",
};

static int
next(struct lexer *lexer, struct location *loc)
{
	int c = getc(lexer->file);
	if (c == '\n') {
		lexer->loc.lineno += 1;
		lexer->loc.colno = 1;
	} else if (c == '\t') {
		lexer->loc.colno += 8;
	} else {
		lexer->loc.colno++;
	}

	if (loc) {
		loc->path = lexer->loc.path;
		loc->lineno = lexer->loc.lineno;
		loc->colno = lexer->loc.colno;
	}

	lexer->cur = c;
	return c;
}

static int
push_back(struct lexer *lexer)
{
	int c = ungetc(lexer->cur, lexer->file);

	if (c == EOF) {
		fatal("failed to ungetc");
	} else if (c == '\n') {
		lexer->loc.lineno -= 1;
	}

	lexer->cur = 0;
	return c;
}

static int
cmp_keyword(const void *lhs, const void *rhs)
{
	return strcmp((char *)lhs, *(const char **)rhs);
}

static enum lex_token
identifier(struct lexer *lexer, struct token *out)
{
	size_t n = 1;
	char buf[256] = {0};
	buf[n - 1] = lexer->cur;

	int c;
	while ((c = next(lexer, NULL)) && (isalnum(c) || c == '_')) {
		buf[n++] = c;
	}

	push_back(lexer);

	void *token = bsearch(buf, tokens, T_LAST_KEYWORD + 1,
			sizeof(tokens[0]), cmp_keyword);

	out->string.data = strdup(buf);
	out->string.n = n;

	if (token) {
		out->type = (const char **)token - tokens;
	} else {
		out->type = T_IDENTIFIER;
	}
	return out->type;
}

static bool
isnumber(int c)
{
	// "xX" is valid hex, "oO" is valid octal, "bB" is valid binary
	return isxdigit(c) || c == 'x' || c == 'X' || c == 'o' || c == 'O'
		|| c == 'b' || c == 'B';
}

static enum lex_token
number(struct lexer *lexer, struct token *out)
{
	size_t n = 1;
	char buf[256] = {0};
	buf[n - 1] = lexer->cur;

	int c;
	while ((c = next(lexer, NULL)) && isnumber(c)) {
		buf[n++] = c;
	}

	out->type = T_NUMBER;

	if (!isspace(c)) {
		snprintf(buf, sizeof(buf), "character '%c' is not a valid number",
			c);
		out->type = T_ERROR;
	}

	out->string.data = strdup(buf);
	out->string.n = n;

	return out->type;
}

static enum lex_token
string(struct lexer *lexer, struct token *out)
{
	while(lexer->cur == '\'') {
		next(lexer, NULL);
	}

	size_t n = 1;
	char buf[PATH_MAX] = {0};
	buf[n - 1] = lexer->cur;

	int c;
	while ((c = next(lexer, NULL)) != '\'') {
		buf[n++] = c;
	}

	out->type = T_STRING;
	out->string.data = strdup(buf);
	out->string.n = n;

	while((c = next(lexer, NULL)) == '\'');

	push_back(lexer);

	return out->type;
}

enum lex_token
lex(struct lexer *lexer, struct token *out)
{
	if (lexer->prev.type != T_ERROR) {
		*out = lexer->prev;
		lexer->prev.type = T_ERROR;
		return out->type;
	}

	int c;
	while ((c = next(lexer, &out->loc)) && isspace(c));

	while (c == '#') {
		while ((c = next(lexer, NULL)) != '\n');
		c = next(lexer, &out->loc);
	}

	if (isalpha(c) || c == '_'){
		return identifier(lexer, out);
	} else if (isdigit(c)) {
		return number(lexer, out);
	}

	switch (c) {
	case '=':
		if ((c = next(lexer, NULL)) == '=') {
			out->type = T_EQ;
		} else {
			out->type = T_ASSIGN;
			push_back(lexer);
		}
		break;
	case ',':
		out->type = T_COMMA;
		break;
	case ':':
		out->type = T_COLON;
		break;
	case '.':
		out->type = T_DOT;
		break;
	case '[':
		out->type = T_LBRACK;
		break;
	case '(':
		out->type = T_LPAREN;
		break;
	case ']':
		out->type = T_RBRACK;
		break;
	case ')':
		out->type = T_RPAREN;
		break;
	case '\'':
		out->type = string(lexer, out);
		break;
	case -1:
	case '\0':
		out->type = T_EOF;
		break;
	default:
		fatal("%s: TODO lex '%c'", loc_str(&out->loc), c);
		break;
	};

	return out->type;
}

enum lex_token
unlex(struct lexer *lexer, struct token *in)
{
	assert(lexer->prev.type == T_ERROR);
	lexer->prev = *in;

	return lexer->prev.type;
}

void
lexer_init(struct lexer *lexer, const char *path)
{
	memset(lexer, 0, sizeof(struct lexer));

	lexer->file = fopen(path, "r");
	if (lexer->file == NULL) {
		fatal("Failed to open file '%s'", path);
	}

	lexer->prev.type = T_ERROR;

	lexer->loc.path = path;
	lexer->loc.lineno = 1;
	lexer->loc.colno = 1;
}

void
lexer_finish(struct lexer *lexer)
{
	fclose(lexer->file);
	memset(lexer, 0, sizeof(struct lexer));
}

void
token_finish(struct token *token)
{
	if (token->type == T_IDENTIFIER || token->type == T_STRING) {
		free(token->string.data);
	}

	token->type = 0;
	token->loc.path = NULL;
	token->loc.lineno = 0;
	token->loc.colno = 0;
}

const char *
token_str(const struct token *token)
{
	static char buf[256] = {0};
	switch(token->type) {
	case T_IDENTIFIER:
		snprintf(buf, sizeof(buf), "identifier '%s'", token->string.data);
		break;
	case T_NUMBER:
		snprintf(buf, sizeof(buf), "number '%s'", token->string.data);
		break;
	case T_STRING:
		snprintf(buf, sizeof(buf), "string '%s'", token->string.data);
		break;
	default:
		return lex_token_str(token->type);
	}

	return buf;
}

const char *
lex_token_str(enum lex_token tok)
{
	switch (tok) {
	case T_EOF:
		return "EOF";
	case T_ERROR:
		return "ERROR";
	default:
		assert(tok < sizeof(tokens) / sizeof(tokens[0]));
		return tokens[tok];
	}
}

const char *
loc_str(const struct location *loc)
{
	static char buf[PATH_MAX] = {0};
	snprintf(buf, sizeof(buf), "%s:%d:%d", loc->path, loc->lineno,
		loc->colno);
	return buf;
}
