#define _XOPEN_SOURCE 700

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include <string.h>

#include "ast.h"
#include "lexer.h"
#include "log.h"
#include "util.h"

#define PATH_MAX 4096

static struct ast_expression *parse_expression(struct lexer *);

static void
unexpected(struct token *token)
{
	if (token->type == T_ERROR) {
		fatal("%s: %s", loc_str(&token->loc), token->string.data);
	} else {
		fatal("%s: unexpected token %s", loc_str(&token->loc),
			token_str(token));
	}
}

static void
expression_list_appened(struct ast_expression_list *list,
		struct ast_expression *expr)
{
	if (list == NULL) {
		fatal("cannot appened expression to empty list");
	}

	const size_t new_size = list->n + 1;

	list->values = xrealloc(list->values,
			new_size * sizeof(struct ast_expression));
	list->values[list->n] = expr;
	list->n = new_size;
}

void
keyword_list_appened(struct ast_keyword_list *list,
		struct ast_identifier *key, struct ast_expression *value)
{
	if (list == NULL) {
		fatal("cannot appened keyword to empty list");
	}

	const size_t new_size = list->n + 1;

	list->keys = xrealloc(list->keys,
			new_size * sizeof(struct ast_identifier));
	list->values = xrealloc(list->values,
			new_size * sizeof(struct ast_expression));
	list->keys[list->n] = key;
	list->values[list->n] = value;
	list->n = new_size;
}

static struct ast_identifier *
parse_identifier(struct token *token)
{
	assert(token->type == T_IDENTIFIER);

	struct ast_identifier *id = xcalloc(1, sizeof(struct ast_identifier));

	id->n = token->string.n;
	id->data = strdup(token->string.data);

	return id;
}

static struct ast_string *
parse_string(struct token *token)
{
	assert(token->type == T_STRING);

	struct ast_string *str = xcalloc(1, sizeof(struct ast_string));

	str->n = token->string.n;
	str->data = strdup(token->string.data);

	return str;
}

static struct ast_expression *parse_primary(struct lexer *);

static struct ast_expression_list *
parse_array(struct lexer *lexer)
{
	struct ast_expression_list *arr = xcalloc(1,
			sizeof(struct ast_expression_list));

	struct token token = {0};
	while (lex(lexer, &token) != T_RBRACK) {
		unlex(lexer, &token);

		struct ast_expression *expr = parse_primary(lexer);

		if (lex(lexer, &token) == T_COMMA) {
			expression_list_appened(arr, expr);
		} else {
			unexpected(&token);
		}
	}

	return arr;
}

static struct ast_number *
parse_number(struct token *token)
{
	assert(token->type == T_NUMBER);

	struct ast_number *num = xcalloc(1, sizeof(struct ast_number));

	const char *buffer = token->string.data;
	const char *data = &buffer[0];
	switch (buffer[1]) {
		case 'x':
		case 'X':
			num->base = NUMBER_HEX;
			break;
		case 'o':
		case 'O':
			data = &buffer[2];
			num->base = NUMBER_OCT;
			break;
		case 'b':
		case 'B':
			data = &buffer[2];
			num->base = NUMBER_BIN;
			break;
		default:
			num->base = NUMBER_DEC;
			break;
	}

	char *endptr = NULL;
	num->value = strtol(data, &endptr, num->base);

	if (*endptr != '\0') {
		fatal("failed to parse number '%s'", token->string.data);
	}


	return num;
}

static struct ast_expression *
parse_primary(struct lexer *lexer)
{
	struct ast_expression *expr = xcalloc(1, sizeof(struct ast_expression));

	struct token token = {0};
	switch (lex(lexer, &token)) {
	case T_IDENTIFIER:
		expr->identifier = parse_identifier(&token);
		expr->type = AST_IDENTIFIER;
		break;
	case T_STRING:
		expr->string = parse_string(&token);
		expr->type = AST_STRING;
		break;
	case T_LBRACK:
		expr->array = parse_array(lexer);
		expr->type = AST_ARRAY;
		break;
	case T_NUMBER:
		expr->number = parse_number(&token);
		expr->type = AST_NUMBER;
		break;
	default:
		unexpected(&token);
		break;
	}

	memcpy(&expr->loc, &token.loc, sizeof(struct location));
	token_finish(&token);

	return expr;
}

static struct ast_arguments *
parse_arguments(struct lexer *lexer)
{
	struct ast_arguments *args = xcalloc(1, sizeof(struct ast_arguments));
	args->position = xcalloc(1, sizeof(struct ast_expression_list));
	args->keyword = xcalloc(1, sizeof(struct ast_keyword_list));

	struct token token = {0};
	while (lex(lexer, &token) != T_RPAREN) {
		unlex(lexer, &token);

		struct ast_expression *expr = parse_expression(lexer);

		switch (lex(lexer, &token)) {
		case T_COMMA:
			expression_list_appened(args->position, expr);
			break;
		case T_COLON:
			if (expr->type != AST_IDENTIFIER) {
				fatal("keyword must be an identifier");
			}
			struct ast_expression *value = parse_expression(lexer);

			keyword_list_appened(args->keyword, expr->identifier,
					value);

			free(expr);
			if (lex(lexer, &token) != T_COMMA) {
				unexpected(&token);
			}

			break;
		case T_RPAREN:
			expression_list_appened(args->position, expr);
			goto argument_out;
			break;
		default:
			unexpected(&token);
			break;
		}
	}

argument_out:
	return args;
}

static struct ast_function *
parse_function(struct lexer *lexer, struct ast_expression *expr)
{
	if (expr->type != AST_IDENTIFIER) {
		fatal("function should be called on an identifier");
	}

	struct ast_function *func = xcalloc(1, sizeof(struct ast_function));

	func->left = expr->identifier;
	func->right = parse_arguments(lexer);

	return func;
}

static struct ast_method *
parse_method(struct lexer *lexer, struct ast_expression *expr)
{
	struct ast_method *method = xcalloc(1, sizeof(struct ast_method));
	struct ast_expression *left = xcalloc(1, sizeof(struct ast_expression));

	left->type = expr->type;
	switch (expr->type) {
	case AST_STRING:
		left->string = expr->string;
		break;
	case AST_IDENTIFIER:
		left->identifier = expr->identifier;
		break;
	default:
		fatal("Syntax error");
		break;
	}

	method->left = left;

	struct ast_expression *func = parse_expression(lexer);
	if (func->type != AST_FUNCTION) {
		fatal("method right side must be a function");
	}

	method->right = func->function;
	free(func);

	return method;
}

static struct ast_expression *
parse_postfix(struct lexer *lexer)
{
	struct ast_expression *expr = parse_primary(lexer);

	struct token token = {0};
	switch (lex(lexer, &token)) {
	case T_LPAREN:
		expr->function = parse_function(lexer, expr);
		expr->type = AST_FUNCTION;
		break;
	case T_DOT:
		expr->method = parse_method(lexer, expr);
		expr->type = AST_METHOD;
		break;
	case T_IDENTIFIER:
		expr->identifier = parse_identifier(&token);
		expr->type = AST_IDENTIFIER;
		break;
	default:
		unlex(lexer, &token);
		break;
	}

	return expr;
}

static struct ast_assignment *
parse_assignment(struct lexer *lexer, struct ast_expression *left)
{
	if (left->type != AST_IDENTIFIER) {
		fatal("%s: assignment target must be an identifier, got %s",
			loc_str(&left->loc), ast_expression_str(left));
	}

	struct ast_assignment *ass = xcalloc(1, sizeof(struct ast_assignment));

	ass->left = left->identifier;

	struct token token = {0};
	switch (lex(lexer, &token)) {
	case T_ASSIGN:
		ass->op = ASSIGNMENT_ASSIGN;
		break;
	case T_STAREQ:
		ass->op = ASSIGNMENT_STAREQ;
		break;
	case T_SLASHEQ:
		ass->op = ASSIGNMENT_SLASHEQ;
		break;
	case T_MODEQ:
		ass->op = ASSIGNMENT_MODEQ;
		break;
	case T_PLUSEQ:
		ass->op = ASSIGNMENT_PLUSEQ;
		break;
	case T_MINEQ:
		ass->op = ASSIGNMENT_MINEQ;
		break;
	default:
		unexpected(&token);
	}

	ass->right = parse_expression(lexer);

	return ass;
}

static struct ast_expression *
parse_expression(struct lexer *lexer)
{
	struct ast_expression *expr = parse_postfix(lexer);

	struct token token = {0};
	switch (lex(lexer, &token)) {
	case T_ASSIGN:
	case T_STAREQ:
	case T_SLASHEQ:
	case T_MODEQ:
	case T_PLUSEQ:
	case T_MINEQ:
		unlex(lexer, &token);
		expr->assignment = parse_assignment(lexer, expr);
		expr->type = AST_ASSIGNMENT;
		break;
	default:
		unlex(lexer, &token);
		break;
	}

	return expr;
}

struct ast_statement *
parse(struct lexer *lexer)
{
	struct ast_statement *root = NULL;

	struct token token = {0};
	while (lex(lexer, &token) != T_EOF) {
		struct ast_statement *statement = xcalloc(1,
				sizeof(struct ast_statement));

		switch (token.type) {
		case T_FOREACH:
			fatal("TODO iteration statement");
			break;
		case T_IF:
			fatal("TODO selection statement");
			break;
		default:
			unlex(lexer, &token);
			statement->expression = parse_expression(lexer);
			statement->type = STATEMENT_EXPRESSION;
			break;
		}

		struct ast_statement **curr = &root;
		while (*curr) {
			curr = &(*curr)->next;
		}
		*curr = statement;
	}

	return root;
}
