#define _XOPEN_SOURCE 700

#include <errno.h>
#include <stddef.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include "getopt_long.h"
#include "interpreter.h"
#include "log.h"
#include "lexer.h"
#include "ninja.h"
#include "setup.h"
#include "util.h"

enum stage {
	STAGE_LEX,
	STAGE_PARSE,
	STAGE_EVAL,
	STAGE_EMIT
};

static enum stage
get_stage(void)
{
	const char *s = getenv("BOSON_STAGE");
	if (!s) {
		return STAGE_EMIT;
	} else if (strcmp(s, "eval") == 0) {
		return STAGE_EVAL;
	} else if (strcmp(s, "parse") == 0) {
		return STAGE_PARSE;
	} else if (strcmp(s, "lex") == 0) {
		return STAGE_LEX;
	} else {
		fatal("unknown BOSON_STAGE value '%s'", s);
	}
	return STAGE_EMIT;
}

static int
setup_usage(void)
{
	printf("usage: boson setup [options] builddir [sourcedir]\n"
			"options:\n"
			"  -h, --help\t\tDisplay this message and exit\n"
			"\n"
			"builddir\t\tDirectory into which the file will be generated, "
			"required\n"
			"sourcedir\t\tDirectory, optional\n"
			"\tDefault to the current working directory\n");
	return 0;
}

int
setup(int argc, char **argv)
{
	if (argc < 2) {
		fprintf(stderr, "Must specify a build directory\n");
		return 1;
	}

	static const struct option options[] = {
		{"help", no_argument, 0, 'h'},
		{0, 0, 0, 0},
	};

	const char *build_dir = NULL;
	const char *source_dir = NULL;

	optind = 1;
	int opt;
	while ((opt = getopt_long(argc, argv, "-h", options, NULL)) != -1) {
		switch (opt) {
		case 'h':
			return setup_usage();
		case '?':
			return 1;
		case 1:
			if (strcmp(optarg, "setup") == 0) {
				break;
			}
			if (build_dir == NULL) {
				build_dir = optarg;
				break;
			}
			if (source_dir == NULL) {
				source_dir = optarg;
				break;
			}
			break;
		default:
			fprintf(stderr, "%s: unrecognized option: %c\n", argv[0], opt);
			return 1;
		}
	}

	info("boson %s", VERSION);

	if (source_dir == NULL) {
		source_dir = ".";
	}

	char *cwd = xcalloc(PATH_MAX, sizeof(char));
	getcwd(cwd, PATH_MAX);

	char abs_source_dir[PATH_MAX] = {0}, abs_build_dir[PATH_MAX] = {0};
	realpath(source_dir, abs_source_dir);
	snprintf(abs_build_dir, PATH_MAX, "%s/%s", cwd, build_dir);
	free(cwd);

	info("source dir: %s", abs_source_dir);
	info("build dir: %s", abs_build_dir);

	char source_path[PATH_MAX] = {0};
	snprintf(source_path, sizeof(source_path), "%s/%s", source_dir,
			"meson.build");

	enum stage stage = get_stage();

	struct lexer lexer = {0};
	lexer_init(&lexer, source_path);

	if (stage == STAGE_LEX) {
		info("printing out tokens");
		struct token token = {0};
		while (lex(&lexer, &token) != T_EOF) {
			info("%s: %s", loc_str(&token.loc), token_str(&token));
			token_finish(&token);
		}
		lexer_finish(&lexer);
		return 0;
	}

	struct ast_statement *root = parse(&lexer);
	lexer_finish(&lexer);

	if (stage == STAGE_PARSE) {
		fatal("TODO print all statements");
	}

	struct context ctx = interpret_ast(root);

	// TODO free statements

	// TODO free ctx
	return emit_ninja(&ctx, abs_build_dir);
}
